#version 430 core

#define MAX_NUM_LIGHTS 16

// transformed vertex attributes
in vec3 worldPos;
in vec3 transNormal;

// directional light
struct DirLight {
  vec3 color;
  vec3 direction;
};

// point lights
struct PtLight {
  vec3 color;
  vec3 position;
  vec3 attenuation;
};

// lights
uniform DirLight dirLights[MAX_NUM_LIGHTS];
uniform PtLight  ptLights [MAX_NUM_LIGHTS];
uniform vec3 ambient;

// number of lights
uniform uint numDirLights;
uniform uint numPtLights;

// camera position
uniform vec3 camPos;

// pixel color
out vec4 FragColor;

// shading constants
const float coeffAmbient  = 0.5f;
const float coeffDiffuse  = 0.8f;
const float coeffSpecular = 0.5f;
const uint  expShininess  = 32;

// draw color
uniform vec4 color;

void main()
{
  // intensities
  vec3 diffuse  = vec3(0.0f);
  vec3 specular = vec3(0.0f);
  float intensity;

  // directions
  vec3 normal  = normalize(transNormal);
  vec3 viewDir = normalize(camPos - worldPos);
  vec3 lightDir, reflectDir;

  // process directional lights
  for (uint l = 0; (l < numDirLights) && (l < MAX_NUM_LIGHTS); l++)
  {
    // calculate directional vectors
    lightDir   = normalize(-dirLights[l].direction);
    reflectDir = normalize(reflect(-lightDir, normal));

    // calculate diffuse component
    intensity = max(dot(normal, lightDir), 0.0f);
    diffuse  += coeffDiffuse * intensity * dirLights[l].color;

    // calculate specular component
    intensity = pow(max(dot(viewDir, reflectDir), 0.0f), expShininess);
    specular += coeffSpecular * intensity * dirLights[l].color;
  }

  // process point lights
  for (uint l = 0; (l < numPtLights) && (l < MAX_NUM_LIGHTS); l++)
  {
    // calculate directional vectors
    lightDir   = normalize(ptLights[l].position - worldPos);
    reflectDir = normalize(reflect(-lightDir, normal));

    // calculate attenuation
    float dist = length(ptLights[l].position - camPos);
    float att = 1.0f / (ptLights[l].attenuation.x +
                        ptLights[l].attenuation.y * dist +
                        ptLights[l].attenuation.z * (dist * dist));

    // calculate diffuse component
    intensity = max(dot(normal, lightDir), 0.0f);
    diffuse  += coeffDiffuse * att * intensity * ptLights[l].color;

    // calculate specular component
    intensity = pow(max(dot(viewDir, reflectDir), 0.0f), expShininess);
    specular += coeffSpecular * att * intensity * ptLights[l].color;
  }

  // sum up components
  vec3 shade = ambient + diffuse + specular;
  FragColor  = color * vec4(shade, 1.0f);

  // normal debugging
  // FragColor = vec4(transNormal, 1.0f);
}
