#version 430 core

// normal image
layout (location = 0) out vec4 FragColor;
// threshold areas
layout (location = 1) out vec4 BrightColor;

#define MAX_NUM_LIGHTS 16

// normal mapping
uniform bool normalMapping;
uniform sampler2D normalMap;
in mat3 TBN;

// transformed vertex attributes
in vec3 worldPos;
in vec3 transNormal;
in vec2 uv;

// directional light
struct DirLight {
  vec3 color;
  vec3 direction;
};

// point lights
struct PtLight {
  vec3 color;
  vec3 position;
  vec3 attenuation;
};

// lights
uniform DirLight dirLights[MAX_NUM_LIGHTS];
uniform PtLight  ptLights [MAX_NUM_LIGHTS];

// lighting
uniform vec3 ambient;
uniform float brightness;

// number of lights
uniform uint numDirLights;
uniform uint numPtLights;

// camera position
uniform vec3 camPos;

// material attributes
uniform sampler2D diffuseColor;
uniform vec3      ambientColor;
uniform vec3      specularColor;
uniform float     expShininess;

void main()
{
  vec3 normal;
  if (normalMapping)
  {
    normal = texture(normalMap, uv).rgb;
    normal = normalize(normal * 2.0 - 1.0);
    normal = normalize(TBN * normal);
  }
  else
  {
    normal = normalize(transNormal);
  }

  // intensities
  vec3 ambientComponent  = ambient * ambientColor;
  vec3 diffuseComponent  = vec3(0.0f);
  vec3 specularComponent = vec3(0.0f);
  float intensity;

  // directions
  vec3 viewDir = normalize(camPos - worldPos);
  vec3 lightDir, reflectDir;

  // process directional lights
  for (uint l = 0; (l < numDirLights) && (l < MAX_NUM_LIGHTS); l++)
  {
    // calculate directional vectors
    lightDir   = normalize(-dirLights[l].direction);
    reflectDir = normalize(reflect(-lightDir, normal));

    // calculate diffuse component
    intensity = max(dot(normal, lightDir), 0.0f);
    diffuseComponent  += intensity * dirLights[l].color;

    // calculate specular component
    intensity = pow(max(dot(viewDir, reflectDir), 0.0f), expShininess);
    specularComponent += intensity * dirLights[l].color;
  }

  // process point lights
  for (uint l = 0; (l < numPtLights) && (l < MAX_NUM_LIGHTS); l++)
  {
    // calculate directional vectors
    lightDir   = normalize(ptLights[l].position - worldPos);
    reflectDir = normalize(reflect(-lightDir, normal));

    // calculate attenuation
    float dist = length(ptLights[l].position - worldPos);
    float att = 1.0f / (ptLights[l].attenuation.x +
                        ptLights[l].attenuation.y * dist +
                        ptLights[l].attenuation.z * (dist * dist));

    // calculate diffuse component
    intensity = max(dot(normal, lightDir), 0.0f);
    diffuseComponent  += att * intensity * ptLights[l].color;
    // calculate specular component

    intensity = pow(max(dot(viewDir, reflectDir), 0.0f), expShininess);
    specularComponent += att * intensity * ptLights[l].color;
  }

  // sample diffuse color, calculate specular color
  vec3 diffColor = vec3(texture(diffuseColor, uv));
  ambientComponent  *= diffColor;
  diffuseComponent  *= diffColor;
  specularComponent *= specularColor;

  // sum up components
  vec3 shade = ambientComponent + diffuseComponent + specularComponent;
  FragColor  = vec4(shade * brightness, 1.0f);

  // normal debugging
  // FragColor = vec4(normal, 1.0f);

  // TBN matrix debugging
  // FragColor = vec4(TBN[0], 1.0f); // T
  // FragColor = vec4(TBN[1], 1.0f); // B
  // FragColor = vec4(TBN[2], 1.0f); // N

  // uv debugging
  // FragColor = vec4(uv.xy, 0.0f, 1.0f);

  float brightness = dot(FragColor.rgb, vec3(0.2126, 0.7152, 0.0722));
  // TODO: tweak threshold
  if(brightness > 0.7)
    BrightColor = vec4(FragColor.rgb, 1.0);
  else
    BrightColor = vec4(0.0, 0.0, 0.0, 1.0);
}
